@@ -23,6 +23,8 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
+#include <stdlib.h>
 #include <sys/types.h>
 
 #define LIBUDEV_I_KNOW_THE_API_IS_SUBJECT_TO_CHANGE
@@ -32,20 +34,22 @@
 
 #include "connman.h"
 
-#ifdef NEED_UDEV_ENUMERATE_ADD_MATCH_PROPERTY
-static int udev_enumerate_add_match_property(struct udev_enumerate *enumerate,
-					const char *property, const char *value)
+#ifdef NEED_UDEV_MONITOR_FILTER
+#if 0
+static int udev_monitor_filter_add_match_subsystem_devtype(struct udev_monitor *udev_monitor,
+				const char *subsystem, const char *devtype)
+{
+	return -EINVAL;
+}
+static int udev_monitor_filter_update(struct udev_monitor *udev_monitor)
+{
+	return -EINVAL;
+}
+static int udev_monitor_filter_remove(struct udev_monitor *udev_monitor)
 {
 	return -EINVAL;
 }
 #endif
-
-#ifdef NEED_UDEV_DEVICE_GET_PARENT_WITH_SUBSYSTEM_DEVTYPE
-static struct udev_device *udev_device_get_parent_with_subsystem_devtype(struct udev_device *device,
-						const char *subsystem, const char *devtype)
-{
-	return NULL;
-}
 #endif
 
 static GSList *device_list = NULL;
@@ -53,19 +57,19 @@ static GSList *device_list = NULL;
 static struct connman_device *find_device(const char *interface)
 {
 	GSList *list;
+	int index;
 
 	if (interface == NULL)
 		return NULL;
 
+	index = connman_inet_ifindex(interface);
+	if (index < 0)
+		return NULL;
+
 	for (list = device_list; list; list = list->next) {
 		struct connman_device *device = list->data;
-		const char *device_interface;
 
-		device_interface = connman_device_get_interface(device);
-		if (device_interface == NULL)
-			continue;
-
-		if (g_str_equal(device_interface, interface) == TRUE)
+		if (connman_device_get_index(device) == index)
 			return device;
 	}
 
@@ -74,20 +78,26 @@ static struct connman_device *find_device(const char *interface)
 
 static void add_device(struct udev_device *udev_device)
 {
-	enum connman_device_type devtype = CONNMAN_DEVICE_TYPE_UNKNOWN;
 	struct connman_device *device;
 	struct udev_list_entry *entry;
-	const char *type = NULL, *interface = NULL;
+	const char *type, *interface = NULL;
+	int index = -1;
 
 	DBG("");
 
+	type = udev_device_get_sysattr_value(udev_device, "type");
+	if (atoi(type) != 1)
+		return;
+
 	entry = udev_device_get_properties_list_entry(udev_device);
 	while (entry) {
 		const char *name = udev_list_entry_get_name(entry);
 
-		if (g_str_has_prefix(name, "CONNMAN_TYPE") == TRUE)
-			type = udev_list_entry_get_value(entry);
-		else if (g_str_has_prefix(name, "CONNMAN_INTERFACE") == TRUE)
+		if (g_str_has_prefix(name, "IFINDEX") == TRUE) {
+			const char *value = udev_list_entry_get_value(entry);
+			if (value != NULL)
+				index = atoi(value);
+		} else if (g_str_has_prefix(name, "INTERFACE") == TRUE)
 			interface = udev_list_entry_get_value(entry);
 
 		entry = udev_list_entry_get_next(entry);
@@ -97,31 +107,13 @@ static void add_device(struct udev_device *udev_device)
 	if (device != NULL)
 		return;
 
-	if (type == NULL || interface == NULL)
+	if (index < 0 || interface == NULL)
 		return;
 
-	if (g_str_equal(interface, "ttyUSB0") == FALSE &&
-				g_str_equal(interface, "noz0") == FALSE)
-		return;
-
-	if (g_str_equal(type, "nozomi") == TRUE)
-		devtype = CONNMAN_DEVICE_TYPE_NOZOMI;
-	else if (g_str_equal(type, "huawei") == TRUE)
-		devtype = CONNMAN_DEVICE_TYPE_HUAWEI;
-	else if (g_str_equal(type, "novatel") == TRUE)
-		devtype = CONNMAN_DEVICE_TYPE_NOVATEL;
-	else
-		return;
-
-	device = connman_device_create(interface, devtype);
+	device = connman_inet_create_device(index);
 	if (device == NULL)
 		return;
 
-	connman_device_set_mode(device, CONNMAN_DEVICE_MODE_NETWORK_SINGLE);
-	connman_device_set_policy(device, CONNMAN_DEVICE_POLICY_MANUAL);
-
-	connman_device_set_interface(device, interface);
-
 	if (connman_device_register(device) < 0) {
 		connman_device_unref(device);
 		return;
@@ -142,7 +134,7 @@ static void remove_device(struct udev_device *udev_device)
 	while (entry) {
 		const char *name = udev_list_entry_get_name(entry);
 
-		if (g_str_has_prefix(name, "CONNMAN_INTERFACE") == TRUE)
+		if (g_str_has_prefix(name, "INTERFACE") == TRUE)
 			interface = udev_list_entry_get_value(entry);
 
 		entry = udev_list_entry_get_next(entry);
@@ -168,9 +160,12 @@ static void print_properties(struct udev_device *device, const char *prefix)
 		const char *value = udev_list_entry_get_value(entry);
 
 		if (g_str_has_prefix(name, "CONNMAN") == TRUE ||
+				g_str_has_prefix(name, "RFKILL") == TRUE ||
 				g_str_has_prefix(name, "ID_MODEM") == TRUE ||
 				g_str_equal(name, "ID_VENDOR") == TRUE ||
 				g_str_equal(name, "ID_MODEL") == TRUE ||
+				g_str_equal(name, "INTERFACE") == TRUE ||
+				g_str_equal(name, "IFINDEX") == TRUE ||
 				g_str_equal(name, "DEVNAME") == TRUE ||
 				g_str_equal(name, "DEVPATH") == TRUE)
 			connman_debug("%s%s = %s", prefix, name, value);
@@ -181,7 +176,7 @@ static void print_properties(struct udev_device *device, const char *prefix)
 
 static void print_device(struct udev_device *device, const char *action)
 {
-	const char *subsystem, *devtype = NULL;
+	const char *subsystem, *sysname, *driver, *devtype = NULL;
 	struct udev_device *parent;
 
 	connman_debug("=== %s ===", action);
@@ -202,6 +197,18 @@ static void print_device(struct udev_device *device, const char *action)
 	parent = udev_device_get_parent_with_subsystem_devtype(device,
 							subsystem, devtype);
 	print_properties(parent, "    ");
+
+	driver = udev_device_get_driver(device);
+	if (driver == NULL) {
+		driver = udev_device_get_driver(parent);
+		if (driver == NULL)
+			return;
+	}
+
+	devtype = udev_device_get_devtype(device);
+	sysname = udev_device_get_sysname(device);
+
+	driver = udev_device_get_driver(parent);
 }
 
 static void enumerate_devices(struct udev *context)
@@ -213,7 +220,7 @@ static void enumerate_devices(struct udev *context)
 	if (enumerate == NULL)
 		return;
 
-	udev_enumerate_add_match_property(enumerate, "CONNMAN_TYPE", "?*");
+	udev_enumerate_add_match_subsystem(enumerate, "net");
 
 	udev_enumerate_scan_devices(enumerate);
 
@@ -241,12 +248,19 @@ static gboolean udev_event(GIOChannel *channel,
 {
 	struct udev_monitor *monitor = user_data;
 	struct udev_device *device;
-	const char *action;
+	const char *subsystem, *action;
 
 	device = udev_monitor_receive_device(monitor);
 	if (device == NULL)
 		return TRUE;
 
+	subsystem = udev_device_get_subsystem(device);
+	if (subsystem == NULL)
+		goto done;
+
+	if (g_str_equal(subsystem, "net") == FALSE)
+		goto done;
+
 	action = udev_device_get_action(device);
 	if (action == NULL)
 		goto done;
@@ -268,6 +282,84 @@ static struct udev *udev_ctx;
 static struct udev_monitor *udev_mon;
 static guint udev_watch = 0;
 
+char *__connman_udev_get_devtype(const char *ifname)
+{
+	struct udev_device *device;
+	const char *devtype;
+	char syspath[128];
+
+	snprintf(syspath, sizeof(syspath) - 1, "/sys/class/net/%s", ifname);
+
+	device = udev_device_new_from_syspath(udev_ctx, syspath);
+	if (device == NULL)
+		return NULL;
+
+	devtype = udev_device_get_devtype(device);
+	if (devtype == NULL)
+		goto done;
+
+done:
+	udev_device_unref(device);
+
+	return NULL;
+}
+
+char *__connman_udev_get_mbm_devnode(const char *ifname)
+{
+	struct udev_device *device, *parent, *control;
+	const char *driver, *devpath1, *devpath2;
+	char *devnode = NULL;
+	unsigned int i;
+
+	device = udev_device_new_from_subsystem_sysname(udev_ctx,
+							"net", ifname);
+	if (device == NULL)
+		return NULL;
+
+	parent = udev_device_get_parent(device);
+	if (parent == NULL)
+		goto done;
+
+	driver = udev_device_get_driver(parent);
+	if (g_strcmp0(driver, "cdc_ether") != 0)
+		goto done;
+
+	parent = udev_device_get_parent_with_subsystem_devtype(device,
+							"usb", "usb_device");
+	if (parent == NULL)
+		goto done;
+
+	devpath1 = udev_device_get_devpath(parent);
+
+	for (i = 0; i < 64; i++) {
+		char sysname[16];
+
+		snprintf(sysname, sizeof(sysname) - 1, "ttyACM%d", i);
+
+		control = udev_device_new_from_subsystem_sysname(udev_ctx,
+							"tty", sysname);
+		if (control == NULL)
+			continue;
+
+		parent = udev_device_get_parent_with_subsystem_devtype(control,
+							"usb", "usb_device");
+		if (parent == NULL)
+			continue;
+
+		devpath2 = udev_device_get_devpath(parent);
+
+		if (g_strcmp0(devpath1, devpath2) == 0) {
+			devnode = g_strdup(udev_device_get_devnode(control));
+			break;
+		}
+	}
+
+done:
+	udev_device_unref(device);
+
+	return devnode;
+}
+
 int __connman_udev_init(void)
 {
 	GIOChannel *channel;
@@ -281,8 +373,7 @@ int __connman_udev_init(void)
 		return -1;
 	}
 
-	udev_mon = udev_monitor_new_from_socket(udev_ctx,
-						"@/org/moblin/connman/udev");
+	udev_mon = udev_monitor_new_from_netlink(udev_ctx, "udev");
 	if (udev_mon == NULL) {
 		connman_error("Failed to create udev monitor");
 		udev_unref(udev_ctx);
